import contextlib
import sys
from abc import ABC
from pathlib import Path
from typing import Any, TypeVar

from transformers.utils import logging
from transformers.utils.generic import ModelOutput
from typing_extensions import TypeVarTuple

from .configuration import RKNNConfig
from .constants import (
    CoreMaskType,
    PlatformType,
)
from .utils.env_utils import get_edge_host_platform
from .utils.import_utils import (
    is_rknn_toolkit_available,
    is_rknn_toolkit_lite_available,
)
from .utils.logging_utils import suppress_output

logger = logging.get_logger(__name__)

if is_rknn_toolkit_lite_available():
    from rknnlite.api import RKNNLite  # pyright: ignore[reportMissingImports]
elif is_rknn_toolkit_available():
    # Fallback to RKNN if RKNNLite is not available. RKNN shares similar functionality as RKNNLite.
    from rknn.api import RKNN as RKNNLite  # pyright: ignore[reportMissingImports]
else:
    logger.error("RKNN Toolkit Lite is not installed. Please install it via pip:")
    logger.error("  pip install rknn-toolkit-lite2==2.3.2")
    sys.exit(-1)


MODEL_OUTPUT_T = TypeVar("MODEL_OUTPUT_T", bound=ModelOutput)
TENSOR_Ts = TypeVarTuple("TENSOR_Ts")


# workaround to enable compatibility between rk-transformers models and transformers pipelines
class PreTrainedModel(ABC):  # noqa: B024
    pass


class RKNNRuntime:
    """Handles RKNN model loading and runtime initialization."""

    def __init__(
        self,
        model_path: str | Path,
        platform: PlatformType | None = None,
        core_mask: CoreMaskType = "auto",
        rknn_config: RKNNConfig | None = None,
    ) -> None:
        self.model_path = Path(model_path)
        self.platform = platform or get_edge_host_platform()
        self.core_mask = core_mask
        self.rknn_config = rknn_config
        self.rknn = self._load_rknn_model()

    def _release(self) -> None:
        """Release RKNNLite resources following RKNN-toolkit2 best practices."""
        if getattr(self, "rknn", None) is not None:
            assert self.rknn is not None
            with contextlib.suppress(Exception), suppress_output():
                self.rknn.release()
            self.rknn = None

    def __del__(self) -> None:  # pragma: no cover - destructor safety
        self._release()

    def list_model_compatible_platform(self) -> dict[str, Any] | None:
        """
        List supported platforms for the loaded model.
        Returns a dictionary of supported platforms or None if the check fails.
        """
        if self.rknn is None:
            return None
        with suppress_output():
            if hasattr(self.rknn, "list_support_target_platform") and callable(self.rknn.list_support_target_platform):  # type: ignore
                return self.rknn.list_support_target_platform(self.model_path.as_posix())  # type: ignore
        return None

    def _load_rknn_model(self) -> RKNNLite:
        """Load the RKNN graph and initialize the runtime."""
        if not self.model_path.exists():
            raise FileNotFoundError(f"RKNN model not found: {self.model_path}")

        # Suppress RKNNLite C-level logging
        with suppress_output():
            rknn = RKNNLite(verbose=False)
            logger.debug("Loading RKNN model from %s", self.model_path)
            ret = rknn.load_rknn(self.model_path.as_posix())
        if ret != 0:
            raise RuntimeError("Failed to load RKNN model")

        # Check model-platform compatibility
        supported_platforms = None
        if hasattr(rknn, "list_support_target_platform") and callable(rknn.list_support_target_platform):  # type: ignore
            with suppress_output():
                supported_platforms: dict[str, Any] | None = rknn.list_support_target_platform(  # type: ignore
                    self.model_path.as_posix()
                )

        if supported_platforms is not None:
            # The return value is an OrderedDict with a key 'support_target_platform' and 'filled_target_platform'
            # containing a list of supported platforms.
            # Note: Only the 'support_target_platform' is relevant for compatibility check.
            support_platforms = [p.lower() for p in supported_platforms.get("support_target_platform", [])]
            if self.platform is not None and len(support_platforms) > 0 and self.platform not in support_platforms:
                raise RuntimeError(
                    f"The model is not compatible with the current platform '{self.platform}'. "
                    f"Supported platforms: {support_platforms}"
                )

        logger.debug(
            "Initializing RKNN runtime (platform=%s, core_mask=%s)",
            self.platform,
            self.core_mask,
        )
        if self.platform in {"rk3588", "rk3576"}:
            core_mask_map = {
                "auto": RKNNLite.NPU_CORE_AUTO,
                "0": RKNNLite.NPU_CORE_0,
                "1": RKNNLite.NPU_CORE_1,
                "2": RKNNLite.NPU_CORE_2,
                "0_1": RKNNLite.NPU_CORE_0_1,
                "0_1_2": RKNNLite.NPU_CORE_0_1_2,
                "all": RKNNLite.NPU_CORE_ALL,
            }
            npu_core = core_mask_map.get(self.core_mask, RKNNLite.NPU_CORE_AUTO)
            with suppress_output():
                ret = rknn.init_runtime(core_mask=npu_core)
                if ret != 0:
                    ret = rknn.init_runtime(target=self.platform, core_mask=npu_core)
        else:
            with suppress_output():
                ret = rknn.init_runtime()

        if ret != 0:
            raise RuntimeError("Failed to initialize RKNN runtime")

        return rknn



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared pytest fixtures for rktransformers tests."""

import shutil
import tempfile
from collections import OrderedDict
from collections.abc import Generator
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from transformers import PretrainedConfig


@pytest.fixture
def temp_dir() -> Generator[Path, None, None]:
    """Create a temporary directory for test files."""
    tmp_dir = Path(tempfile.mkdtemp())
    yield tmp_dir
    if tmp_dir.exists():
        shutil.rmtree(tmp_dir)


@pytest.fixture
def dummy_onnx_file(temp_dir: Path) -> Path:
    """Create a dummy ONNX model file."""
    model_path = temp_dir / "model.onnx"
    model_path.write_text("dummy onnx content")
    return model_path


@pytest.fixture
def dummy_rknn_file(temp_dir: Path) -> Path:
    """Create a dummy RKNN model file."""
    model_path = temp_dir / "model.rknn"
    model_path.touch()
    return model_path


@pytest.fixture
def test_data_dir() -> Path:
    """Return path to the test data directory."""
    return Path(__file__).parent / "data"


@pytest.fixture
def pretrained_config() -> PretrainedConfig:
    """Create a basic pretrained config for testing."""
    return PretrainedConfig(model_type="rknn_model")


@pytest.fixture
def random_bert_model_path(test_data_dir: Path) -> Path:
    """Return path to the random BERT test model."""
    return test_data_dir / "random_bert"


@pytest.fixture
def mock_rknn_lite() -> Generator[MagicMock, None, None]:
    """Mock RKNNLite class and instance."""
    with patch("rktransformers.modeling_utils.RKNNLite") as mock_class:
        mock_instance = mock_class.return_value
        mock_instance.load_rknn.return_value = 0
        mock_instance.init_runtime.return_value = 0
        mock_instance.list_support_target_platform.return_value = OrderedDict(
            [("filled_target_platform", ["rk3588"]), ("support_target_platform", ["rk3588"])]
        )
        yield mock_class



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for RKNN backend integration with CrossEncoder."""

import random
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import torch

from rktransformers import RKCrossEncoder
from rktransformers.utils.env_utils import is_rockchip_platform
from rktransformers.utils.import_utils import (
    is_rknn_toolkit_available,
    is_rknn_toolkit_lite_available,
)


@pytest.mark.requires_rknpu
class TestRKCrossEncoder:
    """Tests for RKNN backend integration with CrossEncoder."""

    def test_rk_cross_encoder(self) -> None:
        """Test using RKCrossEncoder derived class."""
        with (
            patch("rktransformers.integrations.sentence_transformers.load_rknn_model") as mock_load_rknn,
            patch("transformers.AutoConfig.from_pretrained") as mock_config_cls,
            patch("transformers.AutoTokenizer.from_pretrained") as mock_tokenizer_cls,
        ):
            mock_config = MagicMock()
            mock_config.sentence_transformers = {}
            mock_config.sbert_ce_default_activation_function = None
            mock_config.architectures = ["BertForSequenceClassification"]
            mock_config.max_position_embeddings = 512
            mock_config_cls.return_value = mock_config

            mock_tokenizer = MagicMock()
            mock_tokenizer.model_max_length = 512
            mock_tokenizer_cls.return_value = mock_tokenizer

            mock_rknn_model = MagicMock()
            mock_rknn_model.config = mock_config
            mock_load_rknn.return_value = mock_rknn_model

            try:
                model = RKCrossEncoder(
                    "dummy-model",
                    model_kwargs={"file_name": "model.rknn"},
                )
                mock_load_rknn.assert_called_once()
                assert model.backend == "rknn"

            except Exception as e:
                pytest.fail(f"Failed to load mocked RKCrossEncoder: {e}")

    def test_predict_rknn_padding(self) -> None:
        """Test predict method enforces max_length padding."""

        class MockBatchEncoding(dict):
            def to(self, device):
                return self

        with (
            patch("rktransformers.integrations.sentence_transformers.load_rknn_model") as mock_load_rknn,
            patch("transformers.AutoTokenizer.from_pretrained") as mock_tokenizer_cls,
            patch("transformers.AutoConfig.from_pretrained") as mock_config_cls,
        ):
            mock_config = MagicMock()
            mock_config.num_labels = 1
            # Ensure no activation function is accidentally found
            mock_config.sbert_ce_default_activation_function = None
            mock_config.sentence_transformers = {}
            mock_config_cls.return_value = mock_config

            mock_rknn_model = MagicMock()
            mock_rknn_model.config = mock_config  # Link config so it shares the deletions
            mock_rknn_model.config.max_position_embeddings = 128
            mock_rknn_model.device = "cpu"
            mock_rknn_model.return_value = MagicMock(logits=torch.tensor([[0.5]]))  # Mock forward pass output
            mock_load_rknn.return_value = mock_rknn_model

            mock_tokenizer = MagicMock()
            mock_tokenizer.model_max_length = 128
            # Return MockBatchEncoding which has .to() method
            mock_tokenizer.return_value = MockBatchEncoding(
                {"input_ids": torch.tensor([[1, 2]]), "attention_mask": torch.tensor([[1, 1]])}
            )
            mock_tokenizer_cls.return_value = mock_tokenizer

            model = RKCrossEncoder(
                "dummy-model",
                model_kwargs={"file_name": "model.rknn"},
            )

            # Inject _rknn_config to trigger RKNN specific logic in predict
            model._rknn_config = {"batch_size": 1}

            pairs = [["test", "pair"]]
            model.predict(pairs)

            call_args = mock_tokenizer.call_args
            assert call_args is not None
            assert call_args.kwargs.get("padding") == "max_length"
            assert call_args.kwargs.get("max_length") == 128


pytestmark = pytest.mark.skipif(
    not is_rockchip_platform() or (not is_rknn_toolkit_lite_available() and not is_rknn_toolkit_available()),
    reason="Skipping RKNN tests on non-Rockchip platform or missing RKNN Toolkit Lite library.",
)

INTEGRATION_CROSS_ENCODER_MODEL = "rk-transformers/ms-marco-MiniLM-L12-v2"


# RKNN backend gets overloaded when running all tests at once
@pytest.mark.flaky(reruns=1, reruns_delay=random.uniform(15, 30), only_rerun=["RuntimeError"])
@pytest.mark.slow
@pytest.mark.manual
@pytest.mark.integration
@pytest.mark.requires_rknpu
class TestRKCrossEncoderIntegration:
    """Integration tests for RKNN backend with CrossEncoder."""

    @pytest.mark.parametrize(
        "model_id,file_name,batch_size",
        [
            (INTEGRATION_CROSS_ENCODER_MODEL, None, 1),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s512.rknn", 1),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s256.rknn", 1),
            (INTEGRATION_CROSS_ENCODER_MODEL, None, 2),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s512.rknn", 2),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s256.rknn", 2),
            (INTEGRATION_CROSS_ENCODER_MODEL, None, 3),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s512.rknn", 3),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s256.rknn", 3),
            (INTEGRATION_CROSS_ENCODER_MODEL, None, 4),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s512.rknn", 4),
            (INTEGRATION_CROSS_ENCODER_MODEL, "model_b4_s256.rknn", 4),
        ],
    )
    def test_inference_with_file_names_and_batch_sizes(self, model_id: str, file_name: str | None, batch_size: int):
        """Test inference with RKNN backend on real models with different file names and batch sizes.

        Args:
            model_id: Hugging Face model identifier
            file_name: Specific RKNN file to load (or None for default)
            batch_size: Batch size to use for prediction
        """
        from rktransformers import RKCrossEncoder

        # Build model_kwargs with file_name if provided
        model_kwargs = {"platform": "rk3588", "core_mask": "auto"}
        if file_name:
            model_kwargs["file_name"] = file_name

        model = RKCrossEncoder(
            model_id,
            model_kwargs=model_kwargs,
        )

        pairs = [
            ["How old are you?", "What is your age?"],
            ["Hello world", "Hi there!"],
            ["How old are you?", "What is your age?"],
        ]
        scores = model.predict(pairs, batch_size=batch_size)
        assert len(scores) == len(pairs), f"Expected {len(pairs)} scores, got {len(scores)}"
        assert all(isinstance(score, (int, float, np.floating)) for score in scores), "Scores should be numeric"



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration tests for RKNN modeling classes."""

import random

import numpy as np
import pytest
from transformers import AutoTokenizer

from rktransformers.modeling import (
    RKModelForFeatureExtraction,
    RKModelForMaskedLM,
    RKModelForMultipleChoice,
    RKModelForQuestionAnswering,
    RKModelForSequenceClassification,
    RKModelForTokenClassification,
)
from rktransformers.utils.env_utils import is_rockchip_platform
from rktransformers.utils.import_utils import (
    is_rknn_toolkit_available,
    is_rknn_toolkit_lite_available,
)

pytestmark = pytest.mark.skipif(
    not is_rockchip_platform() or (not is_rknn_toolkit_lite_available() and not is_rknn_toolkit_available()),
    reason="Skipping RKNN tests on non-Rockchip platform or missing RKNN Toolkit Lite library.",
)


# RKNN backend gets overloaded when running all tests at once
@pytest.mark.flaky(reruns=1, reruns_delay=random.uniform(15, 30), only_rerun=["RuntimeError"])
@pytest.mark.slow
@pytest.mark.manual
@pytest.mark.integration
@pytest.mark.requires_rknpu
class TestRKModelingIntegration:
    """Integration tests for RKNN modeling classes."""

    def test_feature_extraction(self) -> None:
        """Test RKModelForFeatureExtraction integration."""
        checkpoint = "rk-transformers/all-MiniLM-L6-v2"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForFeatureExtraction.from_pretrained(checkpoint)

        inputs = tokenizer("My name is Philipp and I live in Germany.", return_tensors="np")
        outputs = model(**inputs)
        last_hidden_state = outputs.last_hidden_state
        assert last_hidden_state.shape == (1, 512, 384)  # (batch_size, seq_length, hidden_size)

    def test_masked_lm(self) -> None:
        """Test RKModelForMaskedLM integration."""
        checkpoint = "rk-transformers/bert-base-uncased"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForMaskedLM.from_pretrained(checkpoint)

        inputs = tokenizer("The capital of France is [MASK].", return_tensors="np")
        outputs = model(**inputs)
        logits = outputs.logits
        assert logits.shape == (1, 512, 30522)  # (batch_size, sequence_length, config.vocab_size)

    def test_sequence_classification(self) -> None:
        """Test RKModelForSequenceClassification integration."""
        checkpoint = "rk-transformers/distilbert-base-uncased-finetuned-sst-2-english"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForSequenceClassification.from_pretrained(checkpoint)

        inputs = tokenizer("Hello, my dog is cute", return_tensors="np")
        outputs = model(**inputs)
        logits = outputs.logits
        assert logits.shape == (1, 2)  # (batch_size, config.num_labels)

    def test_question_answering(self) -> None:
        """Test RKModelForQuestionAnswering integration."""
        checkpoint = "rk-transformers/distilbert-base-cased-distilled-squad"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForQuestionAnswering.from_pretrained(checkpoint)

        question, text = "Who was Jim Henson?", "Jim Henson was a nice puppet"
        inputs = tokenizer(question, text, return_tensors="np")
        outputs = model(**inputs)
        start_logits = outputs.start_logits
        end_logits = outputs.end_logits
        assert start_logits.shape == (1, 512)  # (batch_size, sequence_length)
        assert end_logits.shape == (1, 512)  # (batch_size, sequence_length)

    def test_token_classification(self) -> None:
        """Test RKModelForTokenClassification integration."""
        checkpoint = "rk-transformers/bert-base-NER"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForTokenClassification.from_pretrained(checkpoint)

        inputs = tokenizer("My name is Philipp and I live in Germany.", return_tensors="np")
        outputs = model(**inputs)
        logits = outputs.logits
        assert logits.shape == (1, 512, 9)  # (batch_size, sequence_length, config.num_labels)

    def test_multiple_choice(self) -> None:
        """Test RKModelForMultipleChoice integration."""
        checkpoint = "rk-transformers/bert-base-uncased_SWAG"
        tokenizer = AutoTokenizer.from_pretrained(checkpoint)
        model = RKModelForMultipleChoice.from_pretrained(checkpoint)
        # swag dataset and model use 4 choices
        prompt = "In Italy, pizza is served in slices."
        choice0 = "It is eaten with a fork and knife."
        choice1 = "It is eaten while held in the hand."
        choice2 = "It is blended into a smoothie."
        choice3 = "It is folded into a taco."

        encoding = tokenizer(
            [prompt, prompt, prompt, prompt], [choice0, choice1, choice2, choice3], return_tensors="np", padding=True
        )
        inputs = {k: np.expand_dims(v, 0) for k, v in encoding.items()}

        outputs = model(**inputs)
        logits = outputs.logits
        assert logits.shape == (1, 4)  # (batch_size, num_choices)



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for RKNN model loading and platform compatibility."""

import tempfile
from collections import OrderedDict
from pathlib import Path
from typing import cast
from unittest.mock import MagicMock

import numpy as np
import pytest
import torch
from transformers import AutoConfig, PretrainedConfig

from rktransformers.configuration import RKNNConfig
from rktransformers.constants import PlatformType
from rktransformers.modeling import (
    RKModel,
    RKModelForMultipleChoice,
    RKModelForQuestionAnswering,
    RKModelForTokenClassification,
)


class TestRKNNPlatformCheck:
    """Tests for RKNN platform compatibility checking."""

    def test_compatible_platform(
        self, mock_rknn_lite: MagicMock, pretrained_config: PretrainedConfig, dummy_rknn_file: Path
    ) -> None:
        """Test loading model with compatible platform."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform="rk3588")
        assert model.rknn is not None

        supported = model.list_model_compatible_platform()
        assert supported == OrderedDict(
            [("filled_target_platform", ["rk3588"]), ("support_target_platform", ["rk3588"])]
        )

    def test_incompatible_platform(
        self, mock_rknn_lite: MagicMock, pretrained_config: PretrainedConfig, dummy_rknn_file: Path
    ) -> None:
        """Test loading model with incompatible platform raises error."""
        with pytest.raises(RuntimeError, match="not compatible"):
            RKModel(
                config=pretrained_config,
                model_path=dummy_rknn_file,
                platform="rk3566",  # Not in supported list
            )

    def test_case_insensitivity(
        self, mock_rknn_lite: MagicMock, pretrained_config: PretrainedConfig, dummy_rknn_file: Path
    ) -> None:
        """Test that platform matching is case-insensitive."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform="rk3588")
        assert model.rknn is not None

    @pytest.mark.parametrize(
        "platform,should_pass",
        [
            ("rk3588", True),  # Lowercase works
            ("rk3566", False),  # Not supported
            ("rk3568", False),  # Not supported
        ],
    )
    def test_platform_compatibility_variations(
        self,
        mock_rknn_lite: MagicMock,
        pretrained_config: PretrainedConfig,
        dummy_rknn_file: Path,
        platform: str,
        should_pass: bool,
    ) -> None:
        """Test platform compatibility with various platform strings."""
        if should_pass:
            model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform=cast(PlatformType, platform))
            assert model.rknn is not None
        else:
            with pytest.raises(RuntimeError, match="not compatible"):
                RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform=cast(PlatformType, platform))

    def test_multiple_supported_platforms(
        self, mock_rknn_lite: MagicMock, pretrained_config: PretrainedConfig, dummy_rknn_file: Path
    ) -> None:
        """Test model that supports multiple platforms."""
        mock_rknn = mock_rknn_lite.return_value
        mock_rknn.list_support_target_platform.return_value = OrderedDict(
            [
                ("filled_target_platform", ["rk3588", "rk3568"]),
                ("support_target_platform", ["rk3588", "rk3568"]),
            ]
        )

        model1 = RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform="rk3588")
        assert model1.rknn is not None

        model2 = RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform="rk3568")
        assert model2.rknn is not None

        with pytest.raises(RuntimeError, match="not compatible"):
            RKModel(config=pretrained_config, model_path=dummy_rknn_file, platform="rk3566")


class TestRKModelFileName:
    @pytest.fixture
    def temp_model_dir(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)
            # Create dummy config
            config = AutoConfig.from_pretrained("bert-base-uncased")
            config.save_pretrained(tmp_path)

            # Create dummy RKNN files
            (tmp_path / "model.rknn").touch()
            (tmp_path / "model_quantized.rknn").touch()
            (tmp_path / "other.rknn").touch()

            # Create rknn subdirectory with quantized model
            (tmp_path / "rknn").mkdir()
            (tmp_path / "rknn" / "model_w8a8.rknn").touch()

            # Create dummy config.json
            with open(tmp_path / "config.json", "w") as f:
                f.write("{}")

            yield tmp_path

    def test_from_pretrained_with_file_name(self, mock_rknn_lite, temp_model_dir):
        # Test loading specific file
        RKModel.from_pretrained(temp_model_dir, file_name="model_quantized.rknn", platform="rk3588")

        # Verify correct file was loaded
        expected_path = (temp_model_dir / "model_quantized.rknn").as_posix()
        mock_rknn_lite.return_value.load_rknn.assert_called_with(expected_path)

    def test_from_pretrained_default(self, mock_rknn_lite, temp_model_dir):
        # Test loading default (should find model.rknn or first available)
        RKModel.from_pretrained(temp_model_dir, platform="rk3588")

        # Verify it loaded one of the valid files (preference is usually model.rknn)
        # Based on _infer_file_path logic, it prefers RKNN_WEIGHTS_NAME ("model.rknn")
        expected_path = (temp_model_dir / "model.rknn").as_posix()
        mock_rknn_lite.return_value.load_rknn.assert_called_with(expected_path)

    def test_from_pretrained_with_subdirectory_file_name(self, mock_rknn_lite, temp_model_dir):
        # Test loading quantized model from subdirectory
        RKModel.from_pretrained(temp_model_dir, file_name="rknn/model_w8a8.rknn", platform="rk3588")

        # Verify correct file was loaded
        expected_path = (temp_model_dir / "rknn" / "model_w8a8.rknn").as_posix()
        mock_rknn_lite.return_value.load_rknn.assert_called_with(expected_path)

    def test_from_pretrained_file_name_not_found(self, mock_rknn_lite, temp_model_dir):
        # Test loading non-existent file
        with pytest.raises((FileNotFoundError, OSError)):
            RKModel.from_pretrained(temp_model_dir, file_name="non_existent.rknn", platform="rk3588")


class TestRKModelingTasks:
    """Test RKNN model task-specific classes."""

    def test_question_answering(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test RKModelForQuestionAnswering forward pass."""
        mock_rknn = mock_rknn_lite.return_value
        batch_size = 1
        seq_len = 128

        mock_rknn.inference.return_value = [
            np.random.randn(batch_size, seq_len).astype(np.float32),
            np.random.randn(batch_size, seq_len).astype(np.float32),
        ]

        model = RKModelForQuestionAnswering(config=pretrained_config, model_path=dummy_rknn_file)

        input_ids = torch.randint(0, 1000, (batch_size, seq_len))
        attention_mask = torch.ones((batch_size, seq_len))

        outputs = model(input_ids=input_ids, attention_mask=attention_mask)
        assert hasattr(outputs, "start_logits")
        assert hasattr(outputs, "end_logits")
        assert outputs.start_logits.shape == (batch_size, seq_len)
        assert outputs.end_logits.shape == (batch_size, seq_len)

    def test_token_classification(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test RKModelForTokenClassification forward pass."""
        mock_rknn = mock_rknn_lite.return_value
        batch_size = 1
        seq_len = 128
        num_labels = 9

        mock_rknn.inference.return_value = [np.random.randn(batch_size, seq_len, num_labels).astype(np.float32)]

        model = RKModelForTokenClassification(config=pretrained_config, model_path=dummy_rknn_file)

        input_ids = torch.randint(0, 1000, (batch_size, seq_len))
        attention_mask = torch.ones((batch_size, seq_len))

        outputs = model(input_ids=input_ids, attention_mask=attention_mask)
        assert hasattr(outputs, "logits")
        assert outputs.logits.shape == (batch_size, seq_len, num_labels)

    def test_multiple_choice(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test RKModelForMultipleChoice forward pass."""
        mock_rknn = mock_rknn_lite.return_value
        batch_size = 1
        seq_len = 128
        num_choices = 4

        # RKNN returns logits with shape [batch_size, num_choices]
        mock_rknn.inference.return_value = [np.random.randn(batch_size, num_choices).astype(np.float32)]

        # Create config with task_kwargs
        rknn_config = RKNNConfig(task_kwargs={"num_choices": num_choices})
        model = RKModelForMultipleChoice(config=pretrained_config, model_path=dummy_rknn_file, rknn_config=rknn_config)

        input_ids = torch.randint(0, 1000, (batch_size, num_choices, seq_len))
        attention_mask = torch.ones((batch_size, num_choices, seq_len))

        outputs = model(input_ids=input_ids, attention_mask=attention_mask)
        assert hasattr(outputs, "logits")
        assert outputs.logits.shape == (batch_size, num_choices)


class TestInputPaddingMethods:
    """Test input padding and preparation methods."""

    def test_pad_to_model_input_dimensions_2d(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test 2D padding for standard tasks."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        # Test with torch tensor
        input_tensor = torch.randint(0, 1000, (2, 64))
        padded = model._pad_to_model_input_dimensions(input_tensor, padding_id=0, use_torch=True)
        assert padded.shape == (4, 128)

        # Test with numpy array
        input_array = np.random.randint(0, 1000, (2, 64))
        padded = model._pad_to_model_input_dimensions(input_array, padding_id=0, use_torch=False)
        assert padded.shape == (4, 128)

    def test_pad_to_model_input_dimensions_3d(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test 3D padding for multiple-choice tasks."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        # Test with torch tensor
        input_tensor = torch.randint(0, 1000, (2, 4, 64))
        target_shape = (4, 4, 128)
        padded = model._pad_to_model_input_dimensions(
            input_tensor, padding_id=0, use_torch=True, target_shape=target_shape
        )
        assert padded.shape == (4, 4, 128)

        # Test with numpy array
        input_array = np.random.randint(0, 1000, (2, 4, 64))
        padded = model._pad_to_model_input_dimensions(
            input_array, padding_id=0, use_torch=False, target_shape=target_shape
        )
        assert padded.shape == (4, 4, 128)

    def test_pad_to_model_input_dimensions_no_padding_needed(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test padding when input already matches target shape."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        # Input already at target size
        input_tensor = torch.randint(0, 1000, (4, 128))
        padded = model._pad_to_model_input_dimensions(input_tensor, padding_id=0, use_torch=True)
        assert padded.shape == (4, 128)
        assert torch.equal(input_tensor, padded)  # Should be unchanged

    def test_prepare_text_inputs_2d(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test standard 2D input preparation."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        input_ids = torch.randint(0, 1000, (2, 64))
        attention_mask = torch.ones((2, 64))

        use_torch, model_inputs, original_shape = model._prepare_text_inputs(input_ids, attention_mask, None)

        assert use_torch is True
        assert original_shape == (2, 64)
        assert model_inputs["input_ids"].shape == (4, 128)
        assert model_inputs["attention_mask"].shape == (4, 128)

    def test_prepare_text_inputs_3d(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test 3D input preparation for multiple-choice."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        input_ids = torch.randint(0, 1000, (2, 4, 64))
        attention_mask = torch.ones((2, 4, 64))
        target_shape = (4, 4, 128)

        use_torch, model_inputs, original_shape = model._prepare_text_inputs(
            input_ids, attention_mask, None, input_shape=target_shape
        )

        assert use_torch is True
        assert original_shape == (2, 4, 64)
        assert model_inputs["input_ids"].shape == (4, 4, 128)
        assert model_inputs["attention_mask"].shape == (4, 4, 128)

    def test_prepare_text_inputs_with_token_type_ids(self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path) -> None:
        """Test input preparation with token type IDs."""
        # Create config with token types
        config = PretrainedConfig(model_type="rknn_model", type_vocab_size=2)
        model = RKModel(config=config, model_path=dummy_rknn_file, max_seq_length=128, batch_size=4)

        input_ids = torch.randint(0, 1000, (2, 64))
        attention_mask = torch.ones((2, 64))
        token_type_ids = torch.zeros((2, 64))

        use_torch, model_inputs, original_shape = model._prepare_text_inputs(input_ids, attention_mask, token_type_ids)

        assert use_torch is True
        assert original_shape == (2, 64)
        assert model_inputs["input_ids"].shape == (4, 128)
        assert model_inputs["attention_mask"].shape == (4, 128)
        assert model_inputs["token_type_ids"].shape == (4, 128)

    def test_prepare_text_inputs_missing_input_ids(
        self, mock_rknn_lite: MagicMock, dummy_rknn_file: Path, pretrained_config: PretrainedConfig
    ) -> None:
        """Test error handling for missing input_ids."""
        model = RKModel(config=pretrained_config, model_path=dummy_rknn_file)

        with pytest.raises(ValueError, match="`input_ids` is required"):
            model._prepare_text_inputs(None, None, None)



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "0.3.0"


from .configuration import OptimizationConfig, QuantizationConfig, RKNNConfig
from .integrations import RKCrossEncoder, RKSentenceTransformer
from .modeling import (
    RKModel,
    RKModelForFeatureExtraction,
    RKModelForMaskedLM,
    RKModelForMultipleChoice,
    RKModelForQuestionAnswering,
    RKModelForSequenceClassification,
    RKModelForTokenClassification,
)

__all__ = [
    # Config
    "OptimizationConfig",
    "QuantizationConfig",
    "RKNNConfig",
    # Integrations
    "RKCrossEncoder",
    "RKSentenceTransformer",
    # Models
    "RKModel",
    "RKModelForFeatureExtraction",
    "RKModelForMaskedLM",
    "RKModelForMultipleChoice",
    "RKModelForQuestionAnswering",
    "RKModelForSequenceClassification",
    "RKModelForTokenClassification",
]



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import shutil
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Generic, Literal, overload

import numpy as np
import torch
from huggingface_hub import HfApi, ModelHubMixin
from huggingface_hub.constants import HF_HUB_CACHE
from optimum.utils.file_utils import find_files_matching_pattern
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForMaskedLM,
    AutoModelForMultipleChoice,
    AutoModelForQuestionAnswering,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
    AutoTokenizer,
    PretrainedConfig,
)
from transformers.modeling_outputs import (
    BaseModelOutput,
    MaskedLMOutput,
    MultipleChoiceModelOutput,
    QuestionAnsweringModelOutput,
    SequenceClassifierOutput,
    TokenClassifierOutput,
)
from transformers.utils import logging
from transformers.utils.doc import (
    add_end_docstrings,
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
)
from transformers.utils.hub import cached_file, is_offline_mode
from typing_extensions import Unpack

from .configuration import RKNNConfig
from .constants import (
    RKNN_FILE_PATTERN,
    RKNN_WEIGHTS_NAME,
    CoreMaskType,
    PlatformType,
)
from .modeling_utils import MODEL_OUTPUT_T, PreTrainedModel, RKNNRuntime, TENSOR_Ts
from .utils.import_utils import (
    is_rknn_toolkit_lite_available,
)
from .utils.logging_utils import suppress_output

logger = logging.get_logger(__name__)


_TOKENIZER_FOR_DOC = "AutoTokenizer"
RKNN_MODEL_END_DOCSTRING = r"""
    This model inherits from [`~rktransformers.modeling.RKModel`], check its documentation for the generic methods the
    library implements for all its model (such as downloading or saving).
"""

FROM_PRETRAINED_START_DOCSTRING = r"""
    Instantiate a pretrained model from a pre-trained model configuration.

    Args:
        model_id (`Union[str, Path]`):
            Can be either:
                - A string, the *model id* of a pretrained model hosted inside a model repo on huggingface.co.
                    Valid model ids can be located at the root-level, like `bert-base-uncased`, or namespaced under a
                    user or organization name, like `dbmdz/bert-base-german-cased`.
                - A path to a *directory* containing a model saved using [`~OptimizedModel.save_pretrained`],
                    e.g., `./my_model_directory/`.
        export (`bool`, defaults to `False`):
            Defines whether the provided `model_id` needs to be exported to the targeted format.
        force_download (`bool`, defaults to `True`):
            Whether or not to force the (re-)download of the model weights and configuration files, overriding the
            cached versions if they exist.
        token (`Optional[Union[bool,str]]`, defaults to `None`):
            The token to use as HTTP bearer authorization for remote files. If `True`, will use the token generated
            when running `huggingface-cli login` (stored in `huggingface_hub.constants.HF_TOKEN_PATH`).
        cache_dir (`Optional[str]`, defaults to `None`):
            Path to a directory in which a downloaded pretrained model configuration should be cached if the
            standard cache should not be used.
        subfolder (`str`, defaults to `""`):
            In case the relevant files are located inside a subfolder of the model repo either locally or on huggingface.co, you can
            specify the folder name here.
        config (`Optional[transformers.PretrainedConfig]`, defaults to `None`):
            The model configuration.
        local_files_only (`Optional[bool]`, defaults to `False`):
            Whether or not to only look at local files (i.e., do not try to download the model).
        trust_remote_code (`bool`, defaults to `False`):
            Whether or not to allow for custom code defined on the Hub in their own modeling. This option should only be set
            to `True` for repositories you trust and in which you have read the code, as it will execute code present on
            the Hub on your local machine.
        revision (`Optional[str]`, defaults to `None`):
            The specific model version to use. It can be a branch name, a tag name, or a commit id, since we use a
            git-based system for storing models and other artifacts on huggingface.co, so `revision` can be any
            identifier allowed by git.
"""  # noqa: E501

TEXT_INPUTS_DOCSTRING = r"""
    Args:
        input_ids (`Union[torch.Tensor, np.ndarray, None]` of shape `({0})`, defaults to `None`):
            Indices of input sequence tokens in the vocabulary.
            [What are input IDs?](https://huggingface.co/docs/transformers/glossary#input-ids)
        attention_mask (`Union[torch.Tensor, np.ndarray, None]` of shape `({0})`, defaults to `None`):
            Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:
            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        token_type_ids (`Union[torch.Tensor, np.ndarray, None]` of shape `({0})`, defaults to `None`):
            Segment token indices to indicate first and second portions of the inputs.
"""


class RKModel(
    RKNNRuntime,
    PreTrainedModel,
    ModelHubMixin,
    Generic[MODEL_OUTPUT_T, Unpack[TENSOR_Ts]],
    library_name="rk-transformers",
    tags=["rknn", "rockchip", "npu"],
):
    """Base class for RKNN-backed text models integrated with the Hugging Face Hub."""

    model_type: str = "rknn_model"
    auto_model_class = AutoModel

    def __init__(
        self,
        *,
        model_id: str | None = None,
        config: PretrainedConfig | None = None,
        model_path: str | Path,
        platform: PlatformType | None = None,
        core_mask: CoreMaskType = "auto",
        rknn_config: RKNNConfig | None = None,
        max_seq_length: int = 512,
        batch_size: int = 1,
    ) -> None:
        if config is None:
            raise ValueError("A Hugging Face config is required to build an RKModel.")

        super().__init__(model_path=model_path, platform=platform, core_mask=core_mask, rknn_config=rknn_config)
        self.model_id = model_id
        self.config = config

        # Set defaults for input_names, batch_size, and max_seq_length
        self.input_names = ["input_ids", "attention_mask"]
        if getattr(config, "type_vocab_size", 1) > 1:
            self.input_names.append("token_type_ids")
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length

        if self.rknn_config:
            if hasattr(self.rknn_config, "model_input_names") and self.rknn_config.model_input_names:
                self.input_names = self.rknn_config.model_input_names

            if hasattr(self.rknn_config, "max_seq_length") and self.rknn_config.max_seq_length is not None:
                self.max_seq_length = self.rknn_config.max_seq_length

            if hasattr(self.rknn_config, "batch_size"):
                self.batch_size = self.rknn_config.batch_size

        self.pad_token_id = 0
        self.pad_token_type_id = 0
        self.pad_attention_mask = 0  # Huggingface transformers uses 0 for padding attention mask
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
            if hasattr(self.tokenizer, "pad_token_id") and self.tokenizer.pad_token_id is not None:
                self.pad_token_id = self.tokenizer.pad_token_id
            if hasattr(self.tokenizer, "pad_token_type_id") and self.tokenizer.pad_token_type_id is not None:
                self.pad_token_type_id = self.tokenizer.pad_token_type_id
        except Exception:
            logger.warning("Failed to load tokenizer. Using default padding IDs (0).")

        # From optimum.onnxruntime.modeling.ORTModel
        AutoConfig.register(self.model_type, AutoConfig)
        if hasattr(self.auto_model_class, "register"):
            self.auto_model_class.register(AutoConfig, self.__class__)

    @overload
    def __call__(self, *args: Any, return_dict: Literal[False], **kwargs: Any) -> tuple[Unpack[TENSOR_Ts]]: ...  # pyright: ignore[reportOverlappingOverload]

    @overload
    def __call__(self, *args: Any, return_dict: Literal[True], **kwargs: Any) -> MODEL_OUTPUT_T: ...

    @overload
    def __call__(self, *args: Any, **kwargs: Any) -> MODEL_OUTPUT_T: ...

    # return_dict omitted ->  MODEL_OUTPUT_T
    # return_dict=True -> MODEL_OUTPUT_T
    # return_dict=False -> tuple[Unpack[TENSOR_Ts]]
    def __call__(
        self,
        *args: Any,
        return_dict: bool = True,
        **kwargs: Any,
    ) -> MODEL_OUTPUT_T | tuple[Unpack[TENSOR_Ts]]:
        return self.forward(*args, return_dict=return_dict, **kwargs)

    def forward(self, *args: Any, **kwargs: Any) -> MODEL_OUTPUT_T | tuple[Unpack[TENSOR_Ts]]:
        """Define the computation performed at every call.

        Should be overridden by all subclasses.
        """
        raise NotImplementedError(f'Module [{type(self).__name__}] is missing the required "forward" function')

    @property
    def device(self) -> torch.device:
        """Return the device on which the model is stored."""
        return torch.device("cpu")

    def to(self, device: torch.device | str) -> "RKModel":
        """No-op for RKModel. For compatibility with Hugging Face Transformers Pipelines."""
        return self

    def _tensor_to_numpy(self, tensor: torch.Tensor | np.ndarray, dtype: np.dtype[Any]) -> np.ndarray:
        if tensor is None:
            raise ValueError("Input tensor is required for RKNN inference.")
        if isinstance(tensor, torch.Tensor):
            array = tensor.detach().cpu().numpy()
        elif isinstance(tensor, np.ndarray):
            array = tensor
        else:
            array = np.asarray(tensor)
        if array.dtype != dtype:
            array = array.astype(dtype, copy=False)
        return array

    def _torch_if_needed(self, use_torch: bool, array: np.ndarray) -> torch.Tensor | np.ndarray:
        if use_torch:
            contiguous = np.ascontiguousarray(array)
            return torch.from_numpy(contiguous)
        return array

    def _ones_like(self, reference: torch.Tensor | np.ndarray, use_torch: bool) -> torch.Tensor | np.ndarray:
        if use_torch:
            if not isinstance(reference, torch.Tensor):  # pragma: no cover - defensive
                reference = torch.from_numpy(np.asarray(reference))
            return torch.ones_like(reference)
        return np.ones_like(np.asarray(reference))

    def _zeros_like(self, reference: torch.Tensor | np.ndarray, use_torch: bool) -> torch.Tensor | np.ndarray:
        if use_torch:
            if not isinstance(reference, torch.Tensor):  # pragma: no cover - defensive
                reference = torch.from_numpy(np.asarray(reference))
            return torch.zeros_like(reference)
        return np.zeros_like(np.asarray(reference))

    def _pad_to_model_input_dimensions(
        self,
        tensor: torch.Tensor | np.ndarray,
        padding_id: int,
        use_torch: bool,
        target_shape: tuple[int, ...] | None = None,
    ) -> torch.Tensor | np.ndarray:
        """Pad tensor to match model's expected input dimensions.

        Handles arbitrary tensor ranks (2D for standard tasks, 3D for multiple-choice, etc.)
        by padding each dimension independently to match the target shape.

        Args:
            tensor: Input tensor to pad (e.g., shape: [batch, seq_len] or [batch, num_choices, seq_len])
            padding_id: Value to use for padding
            use_torch: Whether to use PyTorch or NumPy for padding
            target_shape: Target shape for the tensor. If None, defaults to 2D padding behavior
                         using self.batch_size and self.max_seq_length.

        Returns:
            Padded tensor with shape matching target_shape
        """
        # Default to 2D padding for backward compatibility
        if target_shape is None:
            target_shape = (getattr(self, "batch_size", tensor.shape[0]), self.max_seq_length)

        if len(target_shape) != len(tensor.shape):
            raise ValueError(
                f"Target shape rank ({len(target_shape)}) must match tensor rank ({len(tensor.shape)}). "
                f"Got target_shape={target_shape}, tensor.shape={tensor.shape}"
            )

        needs_padding = any(current < target for current, target in zip(tensor.shape, target_shape, strict=True))
        if not needs_padding:
            return tensor

        # Calculate padding for each dimension
        # Padding goes at the "end" of each dimension (right/bottom)
        if use_torch:
            # PyTorch pad format: (dim_n_before, dim_n_after, ..., dim_0_before, dim_0_after)
            # We only pad at the end, so all "before" values are 0
            pad_values: list[int] = []
            for current_dim, target_dim in reversed(list(zip(tensor.shape, target_shape, strict=True))):
                pad_values.extend([0, max(0, target_dim - current_dim)])  # (before, after)
            tensor = torch.nn.functional.pad(tensor, tuple(pad_values), value=padding_id)  # type: ignore
        else:
            # NumPy pad format: ((dim_0_before, dim_0_after), (dim_1_before, dim_1_after), ...)
            pad_width = [
                (0, max(0, target - current)) for current, target in zip(tensor.shape, target_shape, strict=True)
            ]
            tensor = np.pad(tensor, pad_width, constant_values=padding_id)

        return tensor

    def _prepare_text_inputs(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None,
        token_type_ids: torch.Tensor | np.ndarray | None,
        input_shape: tuple[int, ...] | None = None,
    ) -> tuple[bool, dict[str, torch.Tensor | np.ndarray | None], tuple[int, ...]]:
        """Prepare text inputs for RKNN inference with padding.

        Args:
            input_ids: Input token IDs
            attention_mask: Attention mask
            token_type_ids: Token type IDs (optional)
            input_shape: Expected input shape (e.g., [batch_size, seq_len] for 2D,
                        [batch_size, num_choices, seq_len] for 3D). If None, defaults
                        to 2D shape using self.batch_size and self.max_seq_length.

        Returns:
            Tuple of (use_torch, model_inputs, original_shape)
        """
        if input_ids is None:
            raise ValueError("`input_ids` is required for RKModel text inference.")

        use_torch = isinstance(input_ids, torch.Tensor)
        original_shape = tuple(input_ids.shape)

        # Calculate target shape
        if input_shape is None:
            # Default 2D behavior: [batch_size, seq_len]
            target_shape = (getattr(self, "batch_size", original_shape[0]), self.max_seq_length)
        else:
            # Use provided input_shape, filling in dimensions as needed
            target_shape = input_shape

        # Pad inputs to target shape
        input_ids = self._pad_to_model_input_dimensions(
            input_ids, padding_id=self.pad_token_id, use_torch=use_torch, target_shape=target_shape
        )

        if attention_mask is None:
            attention_mask = self._ones_like(input_ids, use_torch)
        attention_mask = self._pad_to_model_input_dimensions(
            attention_mask, padding_id=self.pad_attention_mask, use_torch=use_torch, target_shape=target_shape
        )

        if "token_type_ids" in self.input_names:
            if token_type_ids is None:
                token_type_ids = self._zeros_like(input_ids, use_torch)  # Use padded input_ids as reference
            else:
                token_type_ids = self._pad_to_model_input_dimensions(
                    token_type_ids, padding_id=self.pad_token_type_id, use_torch=use_torch, target_shape=target_shape
                )

        return (
            use_torch,
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,
            },
            original_shape,
        )

    def _run_text_model(
        self,
        use_torch: bool,
        model_inputs: dict[str, torch.Tensor | np.ndarray | None],
        expected_outputs: Sequence[str],
    ) -> dict[str, torch.Tensor | np.ndarray]:
        ordered_inputs: list[np.ndarray] = []
        for name in self.input_names:
            tensor = model_inputs.get(name)
            if tensor is None:
                continue
            ordered_inputs.append(self._tensor_to_numpy(tensor, np.dtype(np.int16)))

        if self.rknn is None:
            raise RuntimeError("RKNN runtime has been released and can no longer run inference.")

        # Suppress RKNN inference logs
        with suppress_output():
            if is_rknn_toolkit_lite_available():
                # data_type: int8 | uint8 | int16 | float16 | float32 - limitation with rknn MM API/Hardware
                # This an issue for models with embeddings since they require int64 inputs.
                outputs = self.rknn.inference(inputs=ordered_inputs, data_type="int16")  # type: ignore
            else:
                outputs = self.rknn.inference(inputs=ordered_inputs)
        if outputs is None:
            input_summaries = [f"shape={arr.shape}, dtype={arr.dtype}" for arr in ordered_inputs]
            raise RuntimeError(
                "RKNN inference returned None. "
                "This is likely due to a mismatch between model input shapes and the given inputs. "
                f"Input summary: {input_summaries}"
            )
        if len(outputs) < len(expected_outputs):
            logger.error(
                "RKNN inference output mismatch: expected %d outputs (%s), got %d outputs",
                len(expected_outputs),
                expected_outputs,
                len(outputs),
            )
            raise RuntimeError("RKNN inference did not return the expected number of outputs.")

        prepared: dict[str, torch.Tensor | np.ndarray] = {}
        for idx, name in enumerate(expected_outputs):
            prepared[name] = self._torch_if_needed(use_torch, np.asarray(outputs[idx]))
        return prepared

    def _warn_on_unhandled_inputs(self, kwargs: dict[str, Any]) -> None:
        if kwargs:
            logger.warning_once(  # type: ignore - transformers logger util
                "%s received unsupported arguments: %s",
                self.__class__.__name__,
                ", ".join(kwargs.keys()),
            )

    def _save_pretrained(self, save_directory: Path) -> None:
        target = save_directory / RKNN_WEIGHTS_NAME
        shutil.copyfile(self.model_path, target)

    @staticmethod
    def _cached_file(
        path_or_repo_id: str | Path,
        filename: str,
        subfolder: str = "",
        revision: str | None = "main",
        force_download: bool = False,
        local_files_only: bool = False,
        token: bool | str | None = None,
        cache_dir: str | Path = HF_HUB_CACHE,
        proxies: dict | None = None,
    ) -> Path:
        cached_path = cached_file(
            path_or_repo_id,
            filename=filename,
            subfolder=subfolder,
            revision=revision,
            cache_dir=cache_dir,
            force_download=force_download,
            local_files_only=local_files_only,
            token=token,
            proxies=proxies,
        )
        if cached_path is None:
            raise FileNotFoundError(f"Unable to cache RKNN artifact `{filename}` from {path_or_repo_id}.")
        return Path(cached_path)

    @staticmethod
    def _infer_file_path(
        pattern: str,
        candidate_files: list[Path],
        standard_file_name: str,
        target_file_name: str | None = None,
    ) -> Path:
        if target_file_name is not None:
            specific = [file for file in candidate_files if file.name == target_file_name]
            if not specific:
                raise FileNotFoundError(
                    f"Could not find any RKNN files with target file name {target_file_name}. "
                    f"Candidates: {candidate_files}."
                )
            if len(specific) > 1:
                logger.warning(
                    "Found multiple RKNN files named %s, using %s.",
                    target_file_name,
                    specific[0].name,
                )
            return specific[0]

        standard = [file for file in candidate_files if file.name == standard_file_name]
        if len(standard) == 1:
            return standard[0]
        if len(standard) > 1:
            logger.warning(
                "Found multiple RKNN files named %s, using %s.",
                standard_file_name,
                standard[0].name,
            )
            return standard[0]

        pattern_files = [path for path in candidate_files if re.search(pattern, str(path))]
        if not pattern_files:
            raise FileNotFoundError(
                f"Could not find an RKNN artifact matching pattern {pattern}. Candidates: {candidate_files}."
            )
        if len(pattern_files) > 1:
            logger.warning(
                "Found multiple RKNN files matching pattern %s, using %s.",
                pattern,
                pattern_files[0].name,
            )
        return pattern_files[0]

    @staticmethod
    def _list_repo_rknn_files(
        model_id: str,
        *,
        revision: str | None,
        token: str | bool | None,
        subfolder: str,
    ) -> list[Path]:
        """Enumerate RKNN files from a remote Hugging Face repository."""
        api = HfApi(token=token if isinstance(token, str) else None)
        try:
            repo_files = api.list_repo_files(model_id, revision=revision, repo_type="model")
        except Exception as exc:  # pragma: no cover - network errors
            logger.debug("Failed to list repo files for %s: %s", model_id, exc)
            return []

        filtered: list[Path] = []
        for file_path in repo_files:
            if not re.search(RKNN_FILE_PATTERN, file_path):
                continue
            path_obj = Path(file_path)
            if subfolder:
                try:
                    path_obj.relative_to(subfolder)
                except ValueError:
                    continue
            filtered.append(path_obj)
        return filtered

    @classmethod
    def _resolve_config(
        cls,
        model_id: str,
        config: Any | None,
        *,
        revision: str | None,
        cache_dir: str | Path | None,
        force_download: bool,
        local_files_only: bool,
        token: str | bool | None,
        trust_remote_code: bool,
        proxies: dict | None = None,
    ) -> PretrainedConfig:
        if isinstance(config, PretrainedConfig):
            return config
        if isinstance(config, dict):
            return PretrainedConfig.from_dict(config)
        try:
            return AutoConfig.from_pretrained(
                model_id,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                local_files_only=local_files_only,
                token=token,
                trust_remote_code=trust_remote_code,
                proxies=proxies,
            )
        except Exception as exc:
            logger.warning(
                "Falling back to a generic config for %s because AutoConfig loading failed: %s",
                model_id,
                exc,
            )
            fallback_model_type = getattr(getattr(cls.auto_model_class, "config_class", None), "model_type", None)
            return PretrainedConfig(
                model_type=fallback_model_type or cls.model_type,
                name_or_path=model_id,
            )

    @classmethod
    def _from_pretrained(
        cls,
        *,
        model_id: str,
        config: PretrainedConfig | None,
        # rknn options
        platform: PlatformType | None = None,
        core_mask: CoreMaskType = "auto",
        # hub options
        subfolder: str = "",
        revision: str | None = None,
        force_download: bool = False,
        resume_download: bool | None = False,
        proxies: dict | None = None,
        local_files_only: bool = False,
        trust_remote_code: bool = False,
        cache_dir: str | Path | None,
        token: str | bool | None,
        # file options
        file_name: str | None = None,
        **model_kwargs: Any,
    ):
        cache_dir = cache_dir or HF_HUB_CACHE

        if is_offline_mode() and not local_files_only:
            local_files_only = True

        if os.path.isfile(model_id):
            model_path = Path(model_id)
        elif file_name is not None:
            model_path = cls._cached_file(
                model_id,
                filename=file_name,
                subfolder=subfolder,
                local_files_only=local_files_only,
                force_download=force_download,
                cache_dir=cache_dir,
                revision=revision,
                token=token,
                proxies=proxies,
            )
        else:
            candidate_files = find_files_matching_pattern(
                model_id,
                pattern=RKNN_FILE_PATTERN,
                glob_pattern="**/*.rknn",
                subfolder=subfolder,
                token=token,
                revision=revision,
            )
            if not candidate_files:
                candidate_files = cls._list_repo_rknn_files(
                    model_id,
                    revision=revision,
                    token=token,
                    subfolder=subfolder,
                )
            if not candidate_files:
                raise FileNotFoundError(f"Could not find any RKNN model file in {model_id}.")
            if Path(model_id).is_dir():
                candidate_files = [path.relative_to(model_id) for path in candidate_files]

            resolved_file = cls._infer_file_path(
                RKNN_FILE_PATTERN,
                candidate_files,
                standard_file_name=RKNN_WEIGHTS_NAME,
                target_file_name=file_name,
            )
            subfolder_to_use = resolved_file.parent.as_posix()
            if subfolder_to_use == ".":
                subfolder_to_use = ""

            model_path = cls._cached_file(
                model_id,
                filename=resolved_file.name,
                subfolder=subfolder_to_use,
                local_files_only=local_files_only,
                force_download=force_download,
                cache_dir=cache_dir,
                revision=revision,
                token=token,
                proxies=proxies,
            )

        resolved_config = cls._resolve_config(
            model_id,
            config,
            revision=revision,
            cache_dir=cache_dir,
            force_download=force_download,
            local_files_only=local_files_only,
            token=token,
            trust_remote_code=trust_remote_code,
            proxies=proxies,
        )

        # Try to get rknn config from the resolved config object
        root_rknn_config = {}
        if hasattr(resolved_config, "rknn"):
            root_rknn_config = resolved_config.rknn
        elif isinstance(resolved_config, dict) and "rknn" in resolved_config:
            root_rknn_config = resolved_config["rknn"]

        model_rknn_config = None
        if root_rknn_config:
            # Match model filename to keys in rknn config (e.g. "rknn/model.rknn")
            filename = model_path.name
            for key, conf in root_rknn_config.items():
                if key.endswith(filename):
                    try:
                        model_rknn_config = RKNNConfig.from_dict(conf)
                        logger.info(f"Loaded RKNN config for {filename}")
                        break
                    except Exception as e:
                        logger.warning(f"Failed to parse RKNN config for {key}: {e}")

        if not model_rknn_config:
            logger.warning("RKNN config not found in config.json. Use default batch_size=1 and max_seq_length=512.")

        return cls(
            model_id=model_id,
            config=resolved_config,
            model_path=model_path,
            platform=platform,
            core_mask=core_mask,
            rknn_config=model_rknn_config,
            **model_kwargs,
        )

    @classmethod
    @add_start_docstrings(FROM_PRETRAINED_START_DOCSTRING)
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        *,
        config: PretrainedConfig | None = None,
        # rknn options
        platform: PlatformType | None = None,
        core_mask: CoreMaskType = "auto",
        # hub options
        subfolder: str = "",
        revision: str | None = None,
        force_download: bool = False,
        resume_download: bool | None = False,
        proxies: dict | None = None,
        token: str | bool | None = None,
        local_files_only: bool = False,
        trust_remote_code: bool = False,
        cache_dir: str | Path | None = None,
        # file options
        file_name: str | None = None,
        **model_kwargs: Any,
    ):
        return super().from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            config=config,
            platform=platform,
            core_mask=core_mask,
            subfolder=subfolder,
            revision=revision,
            force_download=force_download,
            proxies=proxies,
            token=token,
            local_files_only=local_files_only,
            trust_remote_code=trust_remote_code,
            cache_dir=cache_dir,
            file_name=file_name,
            **model_kwargs,
        )


FEATURE_EXTRACTION_EXAMPLE = r"""
    Example of feature extraction:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> inputs = tokenizer("My name is Philipp and I live in Germany.", return_tensors="np")

    >>> outputs = model(**inputs)
    >>> last_hidden_state = outputs.last_hidden_state
    >>> list(last_hidden_state.shape)
    [1, 12, 384]
    ```
"""


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForFeatureExtraction(RKModel[BaseModelOutput, torch.Tensor | np.ndarray]):
    """RKNN model for feature extraction tasks."""

    auto_model_class = AutoModel

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, sequence_length")
        + FEATURE_EXTRACTION_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForFeatureExtraction",
            checkpoint="rk-transformers/all-MiniLM-L6-v2",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(input_ids, attention_mask, token_type_ids)
        outputs = self._run_text_model(use_torch, model_inputs, ["last_hidden_state"])
        last_hidden_state = outputs["last_hidden_state"][: original_shape[0]]
        if not return_dict:
            return (last_hidden_state,)
        return BaseModelOutput(last_hidden_state=last_hidden_state)  # type: ignore[arg-type]


MASKED_LM_EXAMPLE = r"""
    Example of masked language modeling:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> inputs = tokenizer("The capital of France is [MASK].", return_tensors="np")

    >>> outputs = model(**inputs)
    >>> logits = outputs.logits
    >>> list(logits.shape)
    [1, 512, 30522]
    ```
"""


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForMaskedLM(RKModel[MaskedLMOutput, torch.Tensor | np.ndarray]):
    """RKNN model for masked language modeling tasks."""

    auto_model_class = AutoModelForMaskedLM

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, sequence_length")
        + MASKED_LM_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForMaskedLM",
            checkpoint="rk-transformers/bert-base-uncased",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(input_ids, attention_mask, token_type_ids)
        outputs = self._run_text_model(use_torch, model_inputs, ["logits"])
        logits = outputs["logits"][: original_shape[0]]
        if not return_dict:
            return (logits,)
        return MaskedLMOutput(logits=logits)  # type: ignore[arg-type]


SEQUENCE_CLASSIFICATION_EXAMPLE = r"""
    Example of single-label classification:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="np")

    >>> outputs = model(**inputs)
    >>> logits = outputs.logits
    >>> list(logits.shape)
    [1, 2]
    ```
"""


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForSequenceClassification(RKModel[SequenceClassifierOutput, torch.Tensor | np.ndarray]):
    """RKNN model for sequence classification/regression tasks."""

    auto_model_class = AutoModelForSequenceClassification

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, sequence_length")
        + SEQUENCE_CLASSIFICATION_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForSequenceClassification",
            checkpoint="rk-transformers/distilbert-base-uncased-finetuned-sst-2-english",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(input_ids, attention_mask, token_type_ids)
        outputs = self._run_text_model(use_torch, model_inputs, ["logits"])
        logits = outputs["logits"][: original_shape[0]]
        if not return_dict:
            return (logits,)
        return SequenceClassifierOutput(logits=logits)  # type: ignore[arg-type]


QUESTION_ANSWERING_EXAMPLE = r"""
    Example of question answering:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> question, text = "Who was Jim Henson?", "Jim Henson was a nice puppet"
    >>> inputs = tokenizer(question, text, return_tensors="np")

    >>> outputs = model(**inputs)
    >>> start_logits = outputs.start_logits
    >>> end_logits = outputs.end_logits
    >>> list(start_logits.shape)
    [1, 512]
    >>> list(end_logits.shape)
    [1, 512]
    ```
"""


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForQuestionAnswering(
    RKModel[QuestionAnsweringModelOutput, torch.Tensor | np.ndarray, torch.Tensor | np.ndarray]
):
    """RKNN Model with a QuestionAnsweringModelOutput for extractive question-answering tasks like SQuAD."""

    auto_model_class = AutoModelForQuestionAnswering

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, sequence_length")
        + QUESTION_ANSWERING_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForQuestionAnswering",
            checkpoint="rk-transformers/distilbert-base-cased-distilled-squad",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(input_ids, attention_mask, token_type_ids)
        outputs = self._run_text_model(use_torch, model_inputs, ["start_logits", "end_logits"])
        start_logits = outputs["start_logits"][: original_shape[0]]
        end_logits = outputs["end_logits"][: original_shape[0]]
        if not return_dict:
            return (start_logits, end_logits)
        return QuestionAnsweringModelOutput(start_logits=start_logits, end_logits=end_logits)  # type: ignore[arg-type]


TOKEN_CLASSIFICATION_EXAMPLE = r"""
    Example of token classification:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> inputs = tokenizer("My name is Philipp and I live in Germany.", return_tensors="np")

    >>> outputs = model(**inputs)
    >>> logits = outputs.logits
    >>> list(logits.shape)
    [1, 512, 9]
    ```
"""


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForTokenClassification(RKModel[TokenClassifierOutput, torch.Tensor | np.ndarray]):
    """RKNN Model with a token classification head on top (a linear layer on top of the hidden-states output)
    e.g. for Named-Entity-Recognition (NER) tasks."""

    auto_model_class = AutoModelForTokenClassification

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, sequence_length")
        + TOKEN_CLASSIFICATION_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForTokenClassification",
            checkpoint="rk-transformers/bert-base-NER",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(input_ids, attention_mask, token_type_ids)
        outputs = self._run_text_model(use_torch, model_inputs, ["logits"])
        logits = outputs["logits"][: original_shape[0]]
        if not return_dict:
            return (logits,)
        return TokenClassifierOutput(logits=logits)  # type: ignore[arg-type]


MULTIPLE_CHOICE_EXAMPLE = r"""
    Example of multiple choice:

    ```python
    >>> from transformers import {processor_class}
    >>> from rktransformers.modeling import {model_class}
    >>> import torch

    >>> tokenizer = {processor_class}.from_pretrained("{checkpoint}")
    >>> model = {model_class}.from_pretrained("{checkpoint}")

    >>> prompt = "In Italy, pizza is served in slices."
    >>> choice0 = "It is eaten with a fork and knife."
    >>> choice1 = "It is eaten while held in the hand."
    >>> choice2 = "It is blended into a smoothie."
    >>> choice3 = "It is folded into a taco."
    >>> labels = torch.tensor(0).unsqueeze(0)  # choice0 is correct (according to Wikipedia ;))

    >>> encoding = tokenizer([prompt, prompt, prompt, prompt], [choice0, choice1, choice2, choice3], return_tensors="np", padding=True)
    >>> inputs = {{k: np.expand_dims(v, 0) for k, v in encoding.items()}}

    >>> outputs = model(**inputs)
    >>> logits = outputs.logits
    >>> list(logits.shape)
    [1, 4]
    ```
"""  # noqa: E501


@add_end_docstrings(RKNN_MODEL_END_DOCSTRING)
class RKModelForMultipleChoice(RKModel[MultipleChoiceModelOutput, torch.Tensor | np.ndarray]):
    """RKNN Model with a multiple choice classification head
    on top (a linear layer on top of the pooled output and a softmax) e.g. for RocStories/SWAG tasks."""

    auto_model_class = AutoModelForMultipleChoice

    @add_start_docstrings_to_model_forward(
        TEXT_INPUTS_DOCSTRING.format("batch_size, num_choices, sequence_length")
        + MULTIPLE_CHOICE_EXAMPLE.format(
            processor_class=_TOKENIZER_FOR_DOC,
            model_class="RKModelForMultipleChoice",
            checkpoint="rk-transformers/bert-base-uncased_SWAG",
        )
    )
    def forward(
        self,
        input_ids: torch.Tensor | np.ndarray | None = None,
        attention_mask: torch.Tensor | np.ndarray | None = None,
        token_type_ids: torch.Tensor | np.ndarray | None = None,
        *,
        return_dict: bool = True,
        **kwargs: Any,
    ):
        self._warn_on_unhandled_inputs(kwargs)
        if input_ids is None:
            raise ValueError("`input_ids` is required for RKModel text inference.")

        # Multiple-choice inputs are 3D: [batch_size, num_choices, seq_len]
        if len(input_ids.shape) != 3:
            raise ValueError(
                f"Multiple-choice inputs must be 3D [batch_size, num_choices, seq_len]. Got shape: {input_ids.shape}"
            )

        batch_size, num_choices, seq_len = input_ids.shape

        # Get num_choices from config if available, otherwise use input shape
        expected_num_choices = None
        if self.rknn_config and hasattr(self.rknn_config, "task_kwargs") and self.rknn_config.task_kwargs:
            expected_num_choices = self.rknn_config.task_kwargs.get("num_choices")
            if expected_num_choices != num_choices:
                raise ValueError(
                    f"Number of choices in config ({expected_num_choices}) does not match input shape ({num_choices})"
                )
        else:
            self.num_choices = num_choices
            logger.warning_once("RKNN config not found in config.json. Using input_ids shape to infer num_choices.")  # type: ignore

        target_shape = (self.batch_size, num_choices, self.max_seq_length)
        use_torch, model_inputs, original_shape = self._prepare_text_inputs(
            input_ids, attention_mask, token_type_ids, input_shape=target_shape
        )
        outputs = self._run_text_model(use_torch, model_inputs, ["logits"])
        logits = outputs["logits"]

        # Reshape logits if needed: RKNN may return [batch_size * num_choices] or [batch_size, num_choices]
        if logits.ndim == 1:
            # Flatten case: [batch_size * num_choices] -> [batch_size, num_choices]
            logits = logits.reshape(original_shape[0], original_shape[1])
        elif logits.shape != (original_shape[0], original_shape[1]):
            # Trim padding if needed
            logits = logits[: original_shape[0], : original_shape[1]]

        if not return_dict:
            return (logits,)
        return MultipleChoiceModelOutput(logits=logits)  # type: ignore[arg-type]



# Copyright 2025 Emmanuel Cortes. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from typing import Any, Literal

import numpy as np
import torch
from sentence_transformers import CrossEncoder, SentenceTransformer
from sentence_transformers.models import Pooling, Transformer
from sentence_transformers.util.file_io import is_sentence_transformer_model
from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging

logger = logging.get_logger(__name__)


def load_rknn_model(model_name_or_path: str, config: PretrainedConfig, task_name: str, **model_kwargs):
    """
    Load an RKNN model using the rktransformers library.

    Args:
        model_name_or_path (str): The model name on Hugging Face or the path to a local model directory.
        config (PretrainedConfig): The model configuration.
        task_name (str): The task name for the model (e.g. 'feature-extraction', 'fill-mask').
        model_kwargs (dict[str, Any]): Additional keyword arguments for the model loading.
            - file_name (str, optional): Specific RKNN file to load (e.g. "model_quantized.rknn").
    """
    try:
        from sentence_transformers.backend.utils import _save_pretrained_wrapper

        from rktransformers.modeling import (
            RKModelForFeatureExtraction,
            RKModelForMaskedLM,
            RKModelForSequenceClassification,
        )

        task_to_model_mapping = {
            "feature-extraction": RKModelForFeatureExtraction,
            "fill-mask": RKModelForMaskedLM,
            "sequence-classification": RKModelForSequenceClassification,
        }

        if task_name not in task_to_model_mapping:
            supported_tasks = ", ".join(task_to_model_mapping.keys())
            raise ValueError(f"Unsupported task: {task_name}. Supported tasks: {supported_tasks}")

        model_cls = task_to_model_mapping[task_name]
    except ImportError as e:
        raise Exception("Using the RKNN backend requires installing the sentence-transformers package.") from e

    model = model_cls.from_pretrained(
        model_name_or_path,
        config=config,
        **model_kwargs,
    )

    # Wrap the save_pretrained method to save the model in the correct subfolder
    # RKNN models are typically just a single .rknn file, but we follow the pattern
    model._save_pretrained = _save_pretrained_wrapper(model._save_pretrained, subfolder="rknn")

    return model


def _load_rknn_config(
    config: PretrainedConfig, model_kwargs: dict[str, Any]
) -> tuple[dict[str, Any] | None, dict[str, Any]]:
    """
    Helper function to load RKNN configuration from config.json.

    Args:
        config (PretrainedConfig): Model configuration
        model_kwargs (dict[str, Any]): Additional model loading arguments

    Returns:
        tuple: (rknn_config dict, updated config, updated model_kwargs dict)
    """
    if not hasattr(config, "rknn"):
        return None, model_kwargs

    root_rknn_config = config.rknn

    file_name = model_kwargs.get("file_name")
    if file_name is None:
        # Preference: model.rknn (unoptimized) -> first available in rknn/
        if "model.rknn" in root_rknn_config:
            file_name = "model.rknn"
        elif root_rknn_config:
            file_name = next(iter(root_rknn_config))

    rknn_config = root_rknn_config.get(file_name)
    if "file_name" not in model_kwargs:
        model_kwargs["file_name"] = file_name

    return rknn_config, model_kwargs


class RKCrossEncoder(CrossEncoder):
    """
    RKNN-compatible CrossEncoder implementation.

    This class extends SentenceTransformers' CrossEncoder to support the RKNN backend.
    It overrides model loading and prediction methods to handle RKNN-specific requirements
    such as fixed input shapes and quantization.
    """

    def __init__(
        self,
        model_name_or_path: str,
        num_labels: int | None = None,
        max_length: int | None = None,
        activation_fn: Any | None = None,
        device: str | None = None,
        cache_folder: str | None = None,
        trust_remote_code: bool = False,
        revision: str | None = None,
        local_files_only: bool = False,
        token: bool | str | None = None,
        model_kwargs: dict | None = None,
        tokenizer_kwargs: dict | None = None,
        config_kwargs: dict | None = None,
        model_card_data: Any | None = None,
        backend: Literal["torch", "onnx", "openvino", "rknn"] = "rknn",
    ) -> None:
        """
        Initialize the RKCrossEncoder.

        Args:
            model_name_or_path (str): The model name on Hugging Face or the path to a local model directory.
            num_labels (int, optional): Number of labels for the classifier.
            max_length (int, optional): Maximum sequence length.
            activation_fn (callable, optional): Activation function.
            device (str, optional): Device to use for computation.
            cache_folder (str, optional): Directory to cache the model files.
            trust_remote_code (bool): Whether to trust remote code.
            revision (str, optional): Model revision to use.
            local_files_only (bool): Whether to only use local files.
            token (bool or str, optional): Hugging Face authentication token.
            model_kwargs (dict, optional): Additional model loading arguments.
            tokenizer_kwargs (dict, optional): Additional tokenizer loading arguments.
            config_kwargs (dict, optional): Additional config loading arguments.
            model_card_data (optional): Model card data.
            backend (str): The backend to use for model loading (default: "rknn").
        """
        super().__init__(
            model_name_or_path,
            num_labels,
            max_length,
            activation_fn,
            device,
            cache_folder,
            trust_remote_code,
            revision,
            local_files_only,
            token,
            model_kwargs,
            tokenizer_kwargs,
            config_kwargs,
            model_card_data,
            backend,  # type: ignore
        )
        # Post-init configuration for RKNN
        if (
            hasattr(self, "_rknn_config")
            and self._rknn_config is not None
            and "max_seq_length" in self._rknn_config
            and hasattr(self, "tokenizer")
            and self.tokenizer is not None
        ):
            self.tokenizer.model_max_length = self._rknn_config["max_seq_length"]

    def _load_model(
        self,
        model_name_or_path: str,
        config: PretrainedConfig,
        backend: str,
        **model_kwargs,
    ) -> None:
        """
        Load the model using the specified backend.

        Overrides CrossEncoder._load_model to support 'rknn' backend.
        """
        if backend == "rknn":
            self._rknn_config, model_kwargs = _load_rknn_config(config, model_kwargs)
            self.model = load_rknn_model(
                model_name_or_path,
                config=config,
                task_name="sequence-classification",
                **model_kwargs,
            )
        else:
            super()._load_model(model_name_or_path, config, backend, **model_kwargs)

    def predict(
        self,
        sentences,
        batch_size: int = 32,
        show_progress_bar: bool | None = None,
        activation_fn: Callable | None = None,
        apply_softmax: bool | None = False,
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
    ) -> list[torch.Tensor] | np.ndarray | torch.Tensor:
        """
        Performs predictions with the CrossEncoder on the given sentence pairs.

        Overrides CrossEncoder.predict to enforce max_length padding for RKNN models.

        Args:
            sentences (Union[List[Tuple[str, str]], Tuple[str, str]]): A list of sentence pairs [(Sent1, Sent2), (Sent3, Sent4)]
                or one sentence pair (Sent1, Sent2).
            batch_size (int, optional): Batch size for encoding. Overridden by RKNN model's configured batch size if set.
            show_progress_bar (bool, optional): Output progress bar. Defaults to None.
            activation_fn (callable, optional): Activation function applied on the logits output of the CrossEncoder.
                If None, the ``model.activation_fn`` will be used, which defaults to :class:`torch.nn.Sigmoid` if num_labels=1, else
                :class:`torch.nn.Identity`. Defaults to None.
            convert_to_numpy (bool, optional): Convert the output to a numpy matrix. Defaults to True.
            apply_softmax (bool, optional): If set to True and `model.num_labels > 1`, applies softmax on the logits
                output such that for each sample, the scores of each class sum to 1. Defaults to False.
            convert_to_numpy (bool, optional): Whether the output should be a list of numpy vectors. If False, output
                a list of PyTorch tensors. Defaults to True.
            convert_to_tensor (bool, optional): Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
                Defaults to False.
        """  # noqa: E501
        if not hasattr(self, "_rknn_config"):
            return super().predict(
                sentences,
                batch_size,
                show_progress_bar,
                activation_fn,
                apply_softmax,
                convert_to_numpy,
                convert_to_tensor,
            )

        from tqdm.autonotebook import trange

        input_was_singular = False
        if sentences and isinstance(sentences, (list, tuple)) and isinstance(sentences[0], str):
            sentences = [sentences]
            input_was_singular = True

        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG  # type: ignore
            )

        if activation_fn is not None:
            self.set_activation_fn(activation_fn, set_default=False)

        # Use RKNN batch size if defined
        if self._rknn_config is not None and "batch_size" in self._rknn_config:
            rknn_batch_size = self._rknn_config["batch_size"]
            if batch_size != rknn_batch_size:
                logger.warning_once(  # type: ignore
                    f"Overriding batch_size {batch_size} with RKNN model's configured batch_size {rknn_batch_size}."
                )
                batch_size = rknn_batch_size

        pred_scores = []
        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            batch = sentences[start_index : start_index + batch_size]
            # Force max_length padding for RKNN
            features = self.tokenizer(
                batch,  # type: ignore
                padding="max_length",
                truncation=True,
                return_tensors="pt",
                max_length=self.max_length,
            )
            model_predictions = self.model(**features, return_dict=True)
            logits = self.activation_fn(model_predictions.logits)

            if apply_softmax and logits.ndim > 1:
                logits = torch.nn.functional.softmax(logits, dim=1)
            pred_scores.extend(logits)

        if self.config.num_labels == 1:
            pred_scores = [score[0] for score in pred_scores]

        if convert_to_tensor:
            pred_scores = torch.stack(pred_scores) if len(pred_scores) else torch.tensor([], device=self.model.device)
        elif convert_to_numpy:
            pred_scores = np.asarray([score.cpu().detach().float().numpy() for score in pred_scores])

        if input_was_singular:
            pred_scores = pred_scores[0]

        return pred_scores


class RKTransformer(Transformer):
    """
    RKNN-compatible Transformer implementation.

    This class extends sentence_transformers.models.Transformer to support the RKNN backend.
    It overrides model loading, initialization, and tokenization methods to handle RKNN-specific
    requirements such as fixed input shapes and quantization.
    """

    def __init__(
        self,
        model_name_or_path: str,
        max_seq_length: int | None = None,
        model_args: dict[str, Any] | None = None,
        tokenizer_args: dict[str, Any] | None = None,
        config_args: dict[str, Any] | None = None,
        cache_dir: str | None = None,
        do_lower_case: bool = False,
        tokenizer_name_or_path: str | None = None,
        backend: str = "rknn",
    ) -> None:
        """
        Initialize the RKTransformer.

        Args:
            model_name_or_path (str): The model name on Hugging Face or the path to a local model directory.
            max_seq_length (int, optional): Maximum sequence length for the model. Can be overridden by
                RKNN configuration.
            model_args (dict, optional): Additional arguments for model loading.
            tokenizer_args (dict, optional): Additional arguments for tokenizer loading.
            config_args (dict, optional): Additional arguments for config loading.
            cache_dir (str, optional): Directory to cache the model files.
            do_lower_case (bool): Whether to lowercase the input text.
            tokenizer_name_or_path (str, optional): The tokenizer name or path to use.
            backend (str): The backend to use for model loading (default: "rknn").
        """
        super().__init__(
            model_name_or_path,
            max_seq_length,
            model_args,
            tokenizer_args,
            config_args,
            cache_dir,
            do_lower_case,
            tokenizer_name_or_path,
            backend,
        )
        # Post-init configuration for RKNN
        if hasattr(self, "_rknn_config") and self._rknn_config is not None and "max_seq_length" in self._rknn_config:
            self.max_seq_length = self._rknn_config["max_seq_length"]

    def _load_model(
        self,
        model_name_or_path: str,
        config,
        cache_dir,
        backend: str,
        is_peft_model: bool,
        **model_kwargs,
    ) -> None:
        """
        Load the model using the specified backend.

        Overrides Transformer._load_model to support 'rknn' backend.

        Args:
            model_name_or_path (str): The model name on Hugging Face or the path to a local model directory.
            config (PretrainedConfig): The model configuration.
            cache_dir (str, optional): Directory to cache the model files.
            backend (str): The backend to use for model loading (e.g., "rknn", "pytorch", "onnx").
            is_peft_model (bool): Whether the model is a PEFT (Parameter-Efficient Fine-Tuning) model.
            **model_kwargs (dict[str, Any]): Additional keyword arguments for model loading, such as:
                - file_name (str, optional): Specific RKNN file to load.
        """
        if backend == "rknn":
            self._rknn_config, model_kwargs = _load_rknn_config(config, model_kwargs)
            self._is_sentence_transformer = is_sentence_transformer_model(model_name_or_path, cache_folder=cache_dir)
            self.auto_model = load_rknn_model(
                model_name_or_path,
                config=config,
                task_name="feature-extraction",
                **model_kwargs,
            )
        else:
            super()._load_model(
                model_name_or_path,
                config,
                cache_dir,
                backend,
                is_peft_model,
                **model_kwargs,
            )

    def tokenize(
        self,
        texts: list[str] | list[dict] | list[tuple[str, str]],
        padding: str | bool = True,
    ) -> dict[str, Any]:
        """
        Tokenize texts with RKNN-specific padding requirements.

        Overrides Transformer.tokenize to enforce max_length padding for RKNN models.

        Args:
            texts (list): A list of texts to tokenize. Can be:
                - list[str]: A list of sentences.
                - list[dict]: A list of dictionaries with text keys.
                - list[tuple[str, str]]: A list of sentence pairs.
            padding (str or bool): Padding strategy. For RKNN models, this is always overridden
                to "max_length" regardless of the provided value.

        Returns:
            dict[str, Any]: A dictionary containing tokenized inputs with keys like 'input_ids',
                'attention_mask', etc.
        """
        if hasattr(self, "_rknn_config"):
            padding = "max_length"
        return super().tokenize(texts, padding)


class RKSentenceTransformer(SentenceTransformer):
    """
    RKNN-compatible SentenceTransformer implementation.

    This class extends SentenceTransformers' SentenceTransformer to support the RKNN backend.
    It overrides the encode method to handle RKNN-specific batch size requirements.
    """

    def __init__(
        self,
        model_name_or_path: str | None = None,
        modules=None,
        device: str | None = None,
        prompts: dict[str, str] | None = None,
        default_prompt_name: str | None = None,
        similarity_fn_name=None,
        cache_folder: str | None = None,
        trust_remote_code: bool = False,
        revision: str | None = None,
        local_files_only: bool = False,
        token=None,
        use_auth_token=None,
        truncate_dim: int | None = None,
        model_kwargs: dict[str, Any] | None = None,
        tokenizer_kwargs: dict[str, Any] | None = None,
        config_kwargs: dict[str, Any] | None = None,
        model_card_data=None,
        backend: str = "rknn",
    ) -> None:
        """
        Initialize the RKSentenceTransformer.

        Args:
            model_name_or_path (str, optional): Model name or path to load.
            backend (str): The backend to use for model loading (default: "rknn").
            All other arguments are passed to SentenceTransformer.__init__.
        """
        super().__init__(
            model_name_or_path,
            modules,
            device,
            prompts,
            default_prompt_name,
            similarity_fn_name,
            cache_folder,
            trust_remote_code,
            revision,
            local_files_only,
            token,
            use_auth_token,
            truncate_dim,
            model_kwargs,
            tokenizer_kwargs,
            config_kwargs,
            model_card_data,
            backend,  # type: ignore
        )

    def _load_auto_model(
        self,
        model_name_or_path: str,
        token: bool | str | None,
        cache_folder: str | None,
        revision: str | None = None,
        trust_remote_code: bool = False,
        local_files_only: bool = False,
        model_kwargs: dict[str, Any] | None = None,
        tokenizer_kwargs: dict[str, Any] | None = None,
        config_kwargs: dict[str, Any] | None = None,
        has_modules: bool = False,
    ) -> list:
        """
        Override _load_auto_model to use RKTransformer for RKNN backend.

        This ensures that when creating models from plain transformers (no modules.json),
        we use RKTransformer which supports the RKNN backend.

        Args:
            model_name_or_path (str): The name or path of the pre-trained model.
            token (Optional[Union[bool, str]]): The token to use for the model.
            cache_folder (Optional[str]): The folder to cache the model.
            revision (Optional[str], optional): The revision of the model. Defaults to None.
            trust_remote_code (bool, optional): Whether to trust remote code. Defaults to False.
            local_files_only (bool, optional): Whether to use only local files. Defaults to False.
            model_kwargs (Optional[Dict[str, Any]], optional): Additional keyword arguments for the model. Defaults to None.
            tokenizer_kwargs (Optional[Dict[str, Any]], optional): Additional keyword arguments for the tokenizer. Defaults to None.
            config_kwargs (Optional[Dict[str, Any]], optional): Additional keyword arguments for the config. Defaults to None.
            has_modules (bool, optional): Whether the model has modules.json. Defaults to False.

        Returns:
            List[RKTransformer | nn.Module]: A list containing the transformer model and the pooling model.
        """  # noqa: E501
        logger.warning(
            f"No sentence-transformers model found with name {model_name_or_path}. "
            "Creating a new one with mean pooling."
        )

        shared_kwargs = {
            "token": token,
            "trust_remote_code": trust_remote_code,
            "revision": revision,
            "local_files_only": local_files_only,
        }
        model_kwargs = shared_kwargs if model_kwargs is None else {**shared_kwargs, **model_kwargs}
        tokenizer_kwargs = shared_kwargs if tokenizer_kwargs is None else {**shared_kwargs, **tokenizer_kwargs}
        config_kwargs = shared_kwargs if config_kwargs is None else {**shared_kwargs, **config_kwargs}

        # Use RKTransformer instead of Transformer
        transformer_model = RKTransformer(
            model_name_or_path,
            cache_dir=cache_folder,
            model_args=model_kwargs,
            tokenizer_args=tokenizer_kwargs,
            config_args=config_kwargs,
            backend=self.backend,
        )
        pooling_model = Pooling(transformer_model.get_word_embedding_dimension(), "mean")
        return [transformer_model, pooling_model]

    def encode(self, sentences: str | list[str], **kwargs):  # type: ignore
        """
        Encode sentences with RKNN backend batch size requirements.

        Overrides SentenceTransformer.encode to enforce the batch size specified in the RKNN
        model configuration when encoding sentences.

        Parameters:
            sentences (str or list of str): The sentence or list of sentences to encode.
            **kwargs: Additional keyword arguments passed to the SentenceTransformer.encode method.
                The 'batch_size' argument will be overridden if the RKNN backend is active.

        Returns:
            np.ndarray: The encoded sentence embeddings as a numpy array.
        """
        batch_size = kwargs.pop("batch_size", 32)

        # Check for RKNN config in modules
        rknn_config: dict | None = None
        for module in self:
            if hasattr(module, "_rknn_config"):
                rknn_config = module._rknn_config  # type: ignore
                break

        if rknn_config is None or "batch_size" not in rknn_config:
            kwargs["batch_size"] = batch_size
            return super().encode(sentences, **kwargs)

        rknn_batch_size = rknn_config["batch_size"]
        if batch_size != rknn_batch_size:
            logger.warning_once(  # type: ignore
                f"Overriding batch_size {batch_size} with RKNN model's configured batch_size {rknn_batch_size}"
            )

        # sentence-transformers models work as usual with adjusted batch size
        is_sentence_transformer = False
        for module in self:
            if hasattr(module, "_is_sentence_transformer"):
                is_sentence_transformer = module._is_sentence_transformer
                break
        if is_sentence_transformer:
            kwargs["batch_size"] = rknn_batch_size
            return super().encode(sentences, **kwargs)

        # For non-sentence-transformers models, batch manually
        if isinstance(sentences, str):
            return super().encode(sentences, **kwargs)

        all_embeddings = []
        for i in range(0, len(sentences), rknn_batch_size):
            batch = sentences[i : i + rknn_batch_size]
            batch_embeddings = super().encode(batch, **kwargs)
            all_embeddings.append(batch_embeddings)

        if not all_embeddings:
            return np.array([])
        return np.vstack(all_embeddings)
